//==============================================================================================================================
// Output Converter 10.0.hlsl
// CricWood "Bieder"  Christopher Maiwald, 20082010
//		E-Mail	CricWood@NerdShack.com
//		ICQ		342 173 470
// Codepage: Windows-1252; text width: 128 chars; tab size: 4 spaces
// File comments are in English language.
//==============================================================================================================================



// Direct3D 10 / 11 difference: Direct3D 10.0 cannot handle multi-sampled textures in the same, easy way Direct3D 11 can.
//	Therefor, multi-sampling is not available in the Direct3D 10 version. However, to avoid two seperate paths in the executable
//	(creating a non-multisampled 2D-texture and a TEX2D SRV vs. creating a multisampled 2D-texture and a TEX2DMS SRV), this D3D
//	10.0 shader declares a probably multi-sampled texture but uses only the first of its samples.
Texture2DMS<float3, 1> g_Input	: register(t0);



cbuffer g_Constants				: register(b0) {
	// The lowest value supported by the display.
	float4	g_Minimum				: packoffset(c0);
	// The highest value supported by the display.
	float4	g_Maximum				: packoffset(c1);
	// "1.0f / (Maximum - Minimum)".
	float3	g_ReciprocalRange		: packoffset(c2);
	// Random numbers as a seed for the dithering's random number generator. Should be between 4 (very poor results are
	//	generated below) and 100k (to avoid integer overruns).
	uint4	g_RandomSeed			: packoffset(c3);
	// Strength of the dithering. For "real" dithering, this is the delta of the display (1/256 with 8 bits per channel). Higher
	//	values result in a noisier image, lower values won't reduce banding.
	float3	g_DitheringStrength		: packoffset(c4);
};



//------------------------------------------------------------------------------------------------------------------------------
// Returns pseudo-random noise in the range (-0.5, 0.5) for the given position.
//------------------------------------------------------------------------------------------------------------------------------
const float3 RGBNoiseFor(
	in const uint2 PixelsPosition
) {
	// A pair of values is created which differs per frame, per texel and per channel:

	// Differs per frame: the random seed.
	// Differs per texel: the linearized texel index.
	// Differs per channel: three constant prime numbers.
	const uint	PixelsIndexBase	= 661 * PixelsPosition.y + PixelsPosition.x;
	const uint3	PixelsIndex		= g_RandomSeed.rgb + PixelsIndexBase * uint3(41, 29, 53);

	// Differs per frame: the random seed.
	// Differs per texel: the number of texels in the rect from (0|0) to "TexelsPosition".
	// Differs per channel: three constant prime numbers.
	const uint	PixelsAreaBase	= PixelsPosition.x * PixelsPosition.y;
	const uint3	PixelsArea		= g_RandomSeed.rgb + PixelsAreaBase * uint3(23, 59, 47);

	// XOR both values for pseudo-random noise. Then move it to (-0.5, +0.5).
	return float3((PixelsIndex ^ PixelsArea) % 661) / 660.0f - 0.5f;
}

//------------------------------------------------------------------------------------------------------------------------------
// The pixel shader. Resolves multi-sampling, applies color space transformations and dithering, and clamps the output.
//------------------------------------------------------------------------------------------------------------------------------
const float3 ProcessPixel(
	in const float4 ItsPosition : SV_Position
) : SV_Target0 {

	// TODO TBD NYI TEMP: Find the pixel's footprint and resolve all of its samples.
	float3 Color = g_Input.Load(int2(ItsPosition.xy), 0);

	// TODO TBD NYI TEMP: Apply color space transformation.
	//	Apply gamma correction.
	Color = pow(saturate(Color), 1.0f / 2.2f);

	const float3 Dithering = g_DitheringStrength * RGBNoiseFor(int2(ItsPosition.xy));
	return clamp(
		(g_Minimum.rgb + Color + Dithering) * g_ReciprocalRange.rgb,
		g_Minimum.rgb, g_Maximum.rgb
	);
}
